@@ -1,4 +1,5 @@
-/* xfce4-newsreader-plugin
+/* 
+ * xfce4-newsreader-plugin
  * A lightweight RSS/Atom news feed reader on xfce4-panel
  * 
  * Copyright 2005 mueki <mueki@users.sourceforge.jp>
@@ -68,11 +69,13 @@ static GtkTargetEntry drop_types[] = {
 };
 
 static gint n_drop_types = sizeof(drop_types) / sizeof(drop_types[0]);
+static const int TOOLTIPS_LEN = 256;
 
 #define BORDER 6
 
 static pthread_mutex_t getfeedlist_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t getfeed_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t msgdisp_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 static feeddata_t*
 create_new_feeddata()
@@ -108,6 +111,7 @@ StoreFeedList2File(newsreader_t* newsreader)
 	gchar* filename;
 
 #if 0
+	/* OPML$B$GJ]B8(B */
 	filename = xfce_resource_save_location (XFCE_RESOURCE_CONFIG,
 											"xfce4" G_DIR_SEPARATOR_S "newsreader.xml",
 											TRUE);
@@ -116,17 +120,16 @@ StoreFeedList2File(newsreader_t* newsreader)
 	g_free(filename); 
 #endif
 
+	/* XBEL$B$GJ]B8(B */
 	filename = xfce_resource_save_location (XFCE_RESOURCE_CONFIG,
 											"xfce4" G_DIR_SEPARATOR_S "newsreader.xml",
 											TRUE);
 	storeFeedList_XBEL(newsreader->poOutlineList, filename);
 	g_free(filename);
 
-
 	return;
 }
 
-
 /* based on gaim_program_is_valid */
 static gboolean
 program_is_valid(const char *program)
@@ -256,15 +259,14 @@ on_itemmenu_activate_cb(GtkMenuItem *item, gpointer data)
 }
 
 /* $B%"%$%3%s%j%9%H$+$i3:Ev$9$k(BURL$B$N(Bfavicon$B$r<hF@(B*/
-static GtkImage*
-getFaviconImage(newsreader_t* newsreader, const gchar* pchUri)
+static GdkPixbuf*
+getFaviconPixbuf(newsreader_t* newsreader, const gchar* pchUri)
 {
 	GList* pIconList = newsreader->poIconList;
-	GtkImage* wImage = GTK_IMAGE(gtk_image_new());
 
 	if(!pchUri){
 		DEBUG_PRINT("uri is null");
-		goto error;
+		return NULL;
 	}
 	
 	for(;pIconList; pIconList = pIconList->next){
@@ -273,22 +275,29 @@ getFaviconImage(newsreader_t* newsreader, const gchar* pchUri)
 			continue;
 		if(strcmp(pIcon->uri, pchUri)==0){
 			if(pIcon->pIconPixbuf){
-				gtk_image_set_from_pixbuf(wImage, gdk_pixbuf_scale_simple(pIcon->pIconPixbuf, 16, 16, GDK_INTERP_BILINEAR));
-				break;
+				return pIcon->pIconPixbuf;
+				/*				gtk_image_set_from_pixbuf(wImage, gdk_pixbuf_scale_simple(pIcon->pIconPixbuf, 16, 16, GDK_INTERP_BILINEAR));
+								break;*/
 			}
 		}
 	}
 
-	if(!pIconList){
-		DEBUG_PRINT("not found icon");
-		goto error;
-	}
-	
-	return wImage;
+	DEBUG_PRINT("not found icon");
+	return NULL;
 
- error:
-	gtk_image_set_from_stock(wImage, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
-	return wImage;
+	/*
+	  if(!pIconList){
+
+	  goto error;
+	  }
+	
+	  return wImage;
+
+	  error:
+	  DEBUG_PRINT("getFaviconImage error");
+	  gtk_image_set_from_stock(wImage, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
+	  return wImage;
+	*/
 }
 
 /* $B%]%C%W%"%C%W%a%K%e!<$N99?7(B */
@@ -344,8 +353,16 @@ UpdateFeedListMenu(newsreader_t* newsreader)
 				pchUri = pFeedData->pFeed->url->str;
 			}
 			
-				
-			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(itmFeed), GTK_WIDGET(getFaviconImage(newsreader, pFeedData->pFeed->url->str)));
+			GdkPixbuf* pIconPixbuf = getFaviconPixbuf(newsreader, pchUri);
+			GtkImage* pIconImage = GTK_IMAGE(gtk_image_new());
+			if(pIconPixbuf){
+				gtk_image_set_from_pixbuf(pIconImage, pIconPixbuf);
+			}
+			else{
+				gtk_image_set_from_stock(pIconImage, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
+			}
+			
+			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(itmFeed), GTK_WIDGET(pIconImage));
 
 			gtk_widget_show(itmFeed);
 			gtk_menu_append(menuFeedList, itmFeed);		
@@ -475,6 +492,11 @@ get_favicon(feed_t* pfeed)
 	GdkPixbuf* pixbuf;
 	xmlURIPtr poURI;
 
+	if(!pfeed){
+		g_warning("feed is null");
+		return NULL;
+	}
+	
 	if(!pfeed->url){
 		g_warning("url is null");
 		return NULL;
@@ -554,6 +576,9 @@ AppendIconList(newsreader_t* newsreader, feeddata_t* pFeedData)
 		favicon = g_new0(favicon_t, 1);
 		favicon->uri = g_strdup(pFeedData->pFeed->url->str);
 		favicon->pIconPixbuf = get_favicon(pFeedData->pFeed);
+		if(favicon->pIconPixbuf){
+			favicon->pIconPixbuf = gdk_pixbuf_scale_simple(favicon->pIconPixbuf, 16, 16, GDK_INTERP_BILINEAR);
+		}
 		newsreader->poIconList = g_list_append(newsreader->poIconList, favicon);
 	}
 
@@ -741,8 +766,8 @@ static gboolean SetTimer (void *p_newsreader)
 	/* Recurrently update the panel-docked monitor through a timer */
 {
     newsreader_t *newsreader = (newsreader_t*) p_newsreader;
-	char buff[256];
-	char dateBuff[256];
+	gchar buff[256];
+	gchar dateBuff[256];
 	item_t* pItem;
 	struct tm *t_st;
 	gint iValidHours;
@@ -793,36 +818,56 @@ static gboolean SetTimer (void *p_newsreader)
 
 	if(pItem->tmModified >= 0){
 		t_st = localtime(&pItem->tmModified);
-		snprintf(dateBuff, sizeof(dateBuff), " - %04d/%02d/%02d %02d:%02d", 
+		snprintf(dateBuff, sizeof(dateBuff), "%04d/%02d/%02d %02d:%02d", 
 				 t_st->tm_year+1900, t_st->tm_mon+1, t_st->tm_mday,
 				 t_st->tm_hour, t_st->tm_min);
+		strncat(buff, " - " , 3);
 		strncat(buff, dateBuff, sizeof(buff));
 	}
 
 	/* set icon */
-	GList* poList = newsreader->poIconList;
 	feed_t* pFeed = ((feeddata_t*)(newsreader->g_pCurFeed->data))->pFeed;
+	gchar* szUrl = NULL;
+	GdkPixbuf* pIconPixbuf = NULL;
 
-	for(; poList; poList = poList->next){
+	if(pFeed && (pFeed->url)){
+		szUrl = pFeed->url->str;
+	}
+
+	if((pIconPixbuf = getFaviconPixbuf(newsreader, szUrl))){
+		gtk_image_set_from_pixbuf(newsreader->wIcon, pIconPixbuf);
+	}
+	else{
+		gtk_image_set_from_stock(newsreader->wIcon, "gtk-dialog-info", GTK_ICON_SIZE_MENU);	
+	}
+	
+	/*	GList* poList = newsreader->poIconList;
+		feed_t* pFeed = ((feeddata_t*)(newsreader->g_pCurFeed->data))->pFeed;
+
+		for(; poList; poList = poList->next){
 		favicon_t* icon = (favicon_t*)poList->data;
 		if(!icon)
-			continue;
+		continue;
 		if(strcmp(icon->uri, pFeed->url->str)==0){
-			if(icon->pIconPixbuf){
-				gtk_image_set_from_pixbuf(newsreader->wIcon, gdk_pixbuf_scale_simple(icon->pIconPixbuf, 16, 16, GDK_INTERP_BILINEAR));
-			}
-			else{
-				gtk_image_set_from_stock(newsreader->wIcon, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
-			}
+		if(icon->pIconPixbuf){
+		gtk_image_set_from_pixbuf(newsreader->wIcon, gdk_pixbuf_scale_simple(icon->pIconPixbuf, 16, 16, GDK_INTERP_BILINEAR));
+		}
+		else{
+		gtk_image_set_from_stock(newsreader->wIcon, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
+		}
 			
-			break;
+		break;
 		}
 
-	}
+		}
 
-	if(!poList){
+		if(!poList){
 		gtk_image_set_from_stock(newsreader->wIcon, "gtk-dialog-info", GTK_ICON_SIZE_MENU);
-	}
+		}
+	*/
+
+	/* set tooltips */
+	gtk_tooltips_set_tip(newsreader->wTooltips, GTK_WIDGET(newsreader->wEvItem), buff, "");
 
 	XFCE_PANEL_UNLOCK();
 
@@ -916,6 +961,16 @@ newsreader_read_config (Control * control, xmlNodePtr node)
 		g_free (value);
 	}
 
+	newsreader->bUseTooltips = FALSE;
+
+	value = xmlGetProp (node, (const xmlChar *) "newsreader_Tooltips");
+    if (value){
+		if(strcmp(value, "true") == 0){
+			newsreader->bUseTooltips = TRUE;
+		} 
+		g_free (value);
+	}
+
 	/* $B%i%Y%k%5%$%:$ND4@0(B */
 	//gtk_widget_set_size_request(GTK_WIDGET(newsreader->wLblItem), newsreader->labelWidth, -1);
     //gtk_widget_set_size_request(newsreader->widget, newsreader->labelWidth, -1);
@@ -964,6 +1019,9 @@ newsreader_write_config (Control * control, xmlNodePtr parent)
 	if(newsreader->pchBrowserCmd)
 		xmlSetProp (parent, "newsreader_BrowserCommand", newsreader->pchBrowserCmd);
 
+	if(newsreader->bUseTooltips)
+		xmlSetProp (parent, "newsreader_Tooltips", "true");
+
 	StoreFeedList2File(newsreader);
 
 	return;
@@ -1058,12 +1116,16 @@ static gboolean
 on_enter_notify_cb (GtkEventBox *widget, GdkEventCrossing *event, newsreader_t* newsreader)
 {
   
-  if(newsreader->iTimerID) {
-	g_source_remove (newsreader->iTimerID);
-	newsreader->iTimerID = 0;
-  }
+	if(newsreader->iTimerID) {
+		g_source_remove (newsreader->iTimerID);
+		newsreader->iTimerID = 0;
+	}
 
-  return TRUE;  
+	if(newsreader->bUseTooltips){
+		gtk_tooltips_enable(newsreader->wTooltips);
+	}
+	
+	return TRUE;  
 }
 
 /* $B%U%)!<%+%9$,30$l$?$H$-$K8F$P$l$F%?%$%^!<$r(B
@@ -1079,9 +1141,66 @@ on_leave_notify_cb (GtkEventBox *widget, GdkEventCrossing *event, newsreader_t*
   newsreader->iTimerID = g_timeout_add (2000,
 										(GtkFunction) SetTimer, newsreader);
 
+  gtk_tooltips_disable(newsreader->wTooltips);
+
   return TRUE;
 }
 
+/* $B%a%C%;!<%8I=<(4X?t(B */
+static gboolean
+msg_display_func(newsreader_t* newsreader)
+{
+	gchar* pchMsg = NULL;
+	GList* pList;
+	msgdata_t* pMsgdata;
+	GtkWidget *dialog;
+
+	pthread_mutex_lock(&msgdisp_mutex);
+		
+	while((pList = newsreader->pMsgList)){
+		if((pMsgdata = (msgdata_t*)pList->data)){
+			if(pMsgdata->pchMsg){
+				pchMsg = g_strdup(pMsgdata->pchMsg);
+				g_free(pMsgdata->pchMsg);
+			}
+			else{
+				DEBUG_PRINT("pMsgdata->pchMsg is null");
+			}
+			newsreader->pMsgList = g_list_remove(newsreader->pMsgList, pMsgdata);
+			g_free(pMsgdata);
+		}
+		else{
+			DEBUG_PRINT("??? pMsgdata is null");
+		}
+		
+		/*xfce_warn(pchMsg);*/
+		DEBUG_PRINT(pchMsg);
+
+		/* FIXME */
+		/* $B0J2<$N%(%i!<$,$G$k$?$a%3%a%s%H%"%&%H(B */
+		/* Xlib: unexpected async reply (sequence 0x178a)!*/
+		/* (xfce4-panel:27398): Gdk-CRITICAL **: gdk_window_set_geometry_hints: assertion `window != NULL' failed*/
+		
+		dialog = gtk_message_dialog_new (NULL,
+										 0, //GTK_DIALOG_DESTROY_WITH_PARENT,
+										 GTK_MESSAGE_WARNING,
+										 GTK_BUTTONS_CLOSE,
+										 (const gchar*)pchMsg);
+
+		/* Destroy the dialog when the user responds to it (e.g. clicks a button) */
+		g_signal_connect_swapped (dialog, "response",
+								  G_CALLBACK (gtk_widget_destroy),
+								  dialog);
+
+		gtk_widget_show_all (dialog);
+		g_free(pchMsg);		
+	}
+	
+	pthread_mutex_unlock(&msgdisp_mutex);
+
+	return FALSE;
+}
+
 /**
  * newsreader_new
  *
@@ -1099,7 +1218,9 @@ newsreader_new (void)
 	GtkWidget *eventbox1;
 
     newsreader = g_new0 (newsreader_t, 1);
-	
+
+	/*	pthread_create(&newsreader->msgdisp_tid, NULL, (void*)msg_display_func, (void *) newsreader);*/
+
     /* initialize default settings */
     newsreader->labelWidth = 300;
 	newsreader->feedUpdateMin = 60;
@@ -1139,9 +1260,15 @@ newsreader_new (void)
     gtk_widget_show (label);
     newsreader->wLblItem = GTK_LABEL(label);
 
+	/* create tooltip */
+	newsreader->wTooltips = gtk_tooltips_new();
+	gtk_tooltips_set_tip(newsreader->wTooltips, GTK_WIDGET(eventbox1), "newsreader", "");
+ 
 	/* create model data*/
 	newsreader->oGUI.wModel =GTK_TREE_MODEL( gtk_list_store_new (NUM_COLS, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT));	
 
+	newsreader->bUseTooltips = TRUE;
+
 	gchar* filename;
 	xmlDocPtr doc = NULL;
 	filename = xfce_resource_lookup (XFCE_RESOURCE_CONFIG,
@@ -1186,6 +1313,14 @@ newsreader_free (Control * control)
 {
     newsreader_t *newsreader = (newsreader_t *) control->data;
 
+	/* stop msg_display_func thread */
+	/*	pthread_mutex_lock(&msgdisp_mutex);
+	newsreader->active = FALSE;
+	pthread_mutex_unlock(&msgdisp_mutex);
+	pthread_cond_signal(&msgdisp_cond);
+	pthread_join(newsreader->msgdisp_tid, NULL);
+	*/
+
     /* free resources, stop timeouts, etc. */
 
 	if(newsreader->iTimerID)
@@ -1312,6 +1447,9 @@ newsreader_apply_options (newsreader_t * newsreader)
 	
 	pango_font_description_free(pFontDesc);
 
+	/* $B%D!<%k%A%C%W(B */
+	newsreader->bUseTooltips = gtk_toggle_button_get_active((GtkToggleButton*)newsreader->oGUI.wChkTooltips);
+
 	gboolean valid;
 	GtkTreeIter iter;
 
@@ -1499,6 +1637,9 @@ newsreader_create_options (Control * control, GtkContainer * container,
 	if(newsreader->pchBrowserCmd)
 		gtk_entry_set_text(GTK_ENTRY(newsreader->oGUI.wEntBrowserCmd), newsreader->pchBrowserCmd);
 
+	/* $B%D!<%k%A%C%W(B */
+	gtk_toggle_button_set_active((GtkToggleButton*)(newsreader->oGUI.wChkTooltips), newsreader->bUseTooltips);
+
 	/* set feed list model data */
 	set_model_data(newsreader);
 
@@ -1594,6 +1735,7 @@ dragedfeed_append_to_list(newsreader_t* newsreader)
 	feed_t* pFeed = NULL;
 	feeddata_t* pFeedData = NULL;
 	outline_t* pOutline = NULL;
+	msgdata_t* pMsgdata;
 
 	pthread_mutex_lock(&getfeed_mutex);
 
@@ -1647,6 +1789,13 @@ dragedfeed_append_to_list(newsreader_t* newsreader)
 
  error:
 	pthread_mutex_unlock(&getfeed_mutex);
+
+	pMsgdata = g_new0(msgdata_t, 1);
+	pMsgdata->pchMsg = g_strdup("Could not Acquire RSS/Atom feed");
+	pthread_mutex_lock(&msgdisp_mutex);
+	newsreader->pMsgList = g_list_append(newsreader->pMsgList, pMsgdata);
+	pthread_mutex_unlock(&msgdisp_mutex);
+	g_idle_add((GSourceFunc)msg_display_func, newsreader);
 	return;
 }
 
@@ -1669,12 +1818,13 @@ void on_wLblItem_drag_drop( GtkWidget        *widget,
 	gchar* pchData = NULL;
 	gchar** ppchSplit = NULL;
 	pthread_t thread;
+	msgdata_t* pMsgdata;
 
 	if ((data->length <= 0) || (data->format != 8)) {
 		g_warning("drag have no data");
 		goto error;
 	}
-	
+
 	len = data->length;
 	p = data->data;
 
@@ -1710,13 +1860,20 @@ void on_wLblItem_drag_drop( GtkWidget        *widget,
 	newsreader->pchURI = g_strdup(pchURI);
 	g_free(pchURI);
 
+	gtk_drag_finish(context, TRUE, TRUE, time_);
+
 	pthread_create(&thread, NULL, (void*)dragedfeed_append_to_list, (void*)newsreader);
 
 	return;
 
  error:
-	gtk_drag_finish(context, FALSE, FALSE, time_);
-	xfce_warn("Could not Acquire RSS/Atom feed");
+	gtk_drag_finish(context, FALSE, TRUE, time_);
+	pMsgdata = g_new0(msgdata_t, 1);
+	pMsgdata->pchMsg = g_strdup("Could not Acquire RSS/Atom feed");
+	pthread_mutex_lock(&msgdisp_mutex);
+	newsreader->pMsgList = g_list_append(newsreader->pMsgList, pMsgdata);
+	pthread_mutex_unlock(&msgdisp_mutex);
+	g_idle_add((GSourceFunc)msg_display_func, newsreader);
 	return;
 }
 
